/*
 * Auto-CreamAPI
 * Copyright (C) 2020 Jeddunk
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see
 * <https://www.gnu.org/licenses/>.
 */

package util;

import com.google.gson.Gson;
import com.google.gson.JsonSyntaxException;
import com.google.gson.reflect.TypeToken;
import kong.unirest.HttpResponse;
import kong.unirest.Unirest;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;
import pojo.App;
import pojo.SteamAppsList;
import util.env.MainEnv;

import java.io.*;
import java.lang.reflect.Type;
import java.time.Duration;
import java.time.Instant;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class SteamAppsListCache {

    private SteamAppsList list = new SteamAppsList();
    private final File cacheFile = new File("steamapps.json");
    private MainEnv env;
    //private String key;

    public SteamAppsListCache() {
        try {
            Class<?> envDefault = Class.forName("util.env.Default");
            env = (MainEnv) envDefault.newInstance();
            // System.out.println(env.getKey());
        } catch (ClassNotFoundException e) {
            env = new MainEnv();
        } catch (IllegalAccessException | InstantiationException e) {
            // Only thrown by newInstance()
            e.printStackTrace();
        }
        //boolean fileFound = true;
        try {
            getListFromFile();
        } catch (FileNotFoundException e) {
            System.err.println("File does not exist, trying to create steamapps.json for the first time...");
            //fileFound = false;
            sync();
        } catch (JsonSyntaxException e) {
            System.err.println("File seems to be corrupt, trying to recreate steamapps.json...");
            //fileFound = false;
            sync();
        }
        if (Instant.now().isAfter(list.getTimestamp().plus(Duration.ofDays(3)))) {
            System.err.println("List in file is not recent!");
            sync();
        }
    }

    private void sync() {
        getListFromApi();
        try {
            saveListToFile();
        } catch (IOException ex) {
            ex.printStackTrace();
        }
    }

    public App getGame(int appId) {
        return list.getSteamAppsList().stream().filter(app -> app.getAppId() == appId).findFirst().orElse(null);
    }

    @SuppressWarnings("unused")
    public App getGame(String name) {
        return list.getSteamAppsList().stream()
                .filter(app -> app.getName().equalsIgnoreCase(name)).findFirst().orElse(null);
    }

    public App findGame(String name) {
        for (App app : list.getSteamAppsList()) {
            if (app.getName().toLowerCase().replaceAll("[^a-zA-Z0-9\\s+]", "")
                    .startsWith(name.toLowerCase().replaceAll("[^a-zA-Z0-9\\s+]", ""))) {
                return app;
            }
        }
        for (App app : list.getSteamAppsList()) {
            if (app.getName().toLowerCase().replaceAll("[^a-zA-Z0-9\\s+]", "")
                    .contains(name.toLowerCase().replaceAll("[^a-zA-Z0-9\\s+]", ""))) {
                return app;
            }
        }
        return null;
    }

    private void getListFromApi() {
        System.out.println("Trying to get SteamAppList from API...");
        List<App> apps = getApps();
        list.setTimestamp(Instant.now());
        list.setSteamAppsList(apps);
    }

    private List<App> getApps() {
        HttpResponse<String> httpResponse =
                Unirest.get("https://api.steampowered.com/IStoreService/GetAppList/v1/" +
                        "?key=" + env.getKey() + "&include_games=1&max_results=50000").asString();
        return new Gson()
                .fromJson(httpResponse.getBody(), App.Rezponze.Download.class)
                .getResponse().getApps();
    }

    private void saveListToFile() throws IOException {
        System.out.println("Trying to save SteamAppList to file (" + cacheFile.getAbsolutePath() + ")...");
        Gson gson = new Gson();
        String jsonString = gson.toJson(list);
        BufferedWriter fOut = new BufferedWriter(new FileWriter(cacheFile));
        fOut.write(jsonString);
        fOut.close();
        System.out.println("Successfully saved SteamAppList to file  (" + cacheFile.getAbsolutePath() + ")...");
    }

    private void getListFromFile() throws FileNotFoundException, JsonSyntaxException {
        System.out.println("Trying to get SteamAppList from file (" + cacheFile.getAbsolutePath() + ")...");
        BufferedReader fIn = new BufferedReader(new FileReader(cacheFile));
        String json = fIn.lines().collect(Collectors.joining());
        final Type type = new TypeToken<SteamAppsList>() {
        }.getType();
        Gson gson = new Gson();
        list = gson.fromJson(json, type);
        System.out.println("Successfully got SteamAppList from file (" + cacheFile.getAbsolutePath() + ")...");
    }

    public LinkedHashMap<Number, String> getDlcMap(String appId) throws IOException{
        Map<Integer, String> steamStoreDLCs = new HashMap<>();
        Map<Integer, String> steamDbDLCs = new HashMap<>();
        //StringBuilder sb = new StringBuilder();
        try {
            // Steam Store
            Document steamDoc = Jsoup
                    .connect("https://store.steampowered.com/app/" + appId + "/")
                    .get();
            Elements steamDLCs = steamDoc.getElementsByClass("game_area_dlc_row");
            for (Element dlc : steamDLCs) {
                String dlc_id = dlc.attr("data-ds-appid");
                String dlc_name = dlc
                        .getElementsByClass("game_area_dlc_name")
                        .text().replace("\n", "").trim();
                steamStoreDLCs.put(Integer.parseInt(dlc_id), dlc_name);
            }
            // SteamDB
            Document steamDbDoc = Jsoup
                    .connect("https://steamdb.info/app/" + appId + "/dlc/")
                    .userAgent("Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:76.0) Gecko/20100101 Firefox/76.0")
                    .get();
            Element steamDbDlcSection = steamDbDoc.getElementById("dlc");
            Elements steamDbDLCElements = steamDbDlcSection.getElementsByClass("app");
            for (Element dlc : steamDbDLCElements) {
                String dlc_id = dlc.attr("data-appid");
                String dlc_name = "Unknown DLC " + dlc_id;
                Elements td = dlc.getElementsByTag("td");
                if (!td.isEmpty()) {
                    dlc_name = td.get(1).text().replace("\n", "").trim();
                }
                steamDbDLCs.put(Integer.parseInt(dlc_id), dlc_name);
            }
        } catch (NullPointerException e) {
            // ignore
        }

        Map<Integer, String> allDLCs = new HashMap<>(steamStoreDLCs);
        steamDbDLCs.forEach(allDLCs::putIfAbsent);
        return allDLCs.entrySet().stream()
                .sorted(Map.Entry.comparingByKey())
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue,
                        (oldValue, newValue) -> oldValue, LinkedHashMap::new));
    }
}
